// ###################################################################################
// This program retrieves air quality measures from Air-Sensor USB-stick CO-20
//
// Compile with:  gcc -o airsensor airsensor.c -lusb
// Tested on Raspberry Pi (jessie)
// Modified code by PumpkinEater (2016)
// 
// This code is based on the following source code (Airsensor and TCP client):
//  Airsensor:
//   Original source: Rodric Yates http://code.google.com/p/airsensor-linux-usb/
//   Modified from source: Ap15e (MiOS) http://wiki.micasaverde.com/index.php/CO2_Sensor
//  TCP client:
//   Author: dreamshader, http://www.forum-raspberrypi.de/Thread-c-c-ein-einfacher-tcp-ip-client-fuer-den-rpi)    
// ###################################################################################

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#include <getopt.h>
#include <signal.h>
#include <assert.h>
#include <usb.h>
#include <asm/byteorder.h>
#include <sys/wait.h>
#include <time.h>

struct usb_dev_handle *devh;

/* -----------------------------------------------------------------------------------------*/
void getArguments ( int argc, char **argv, int *pPort, char **pHostname, char **pSensorname, int *pIntervall )
{
  int next_option;

  /* valid short options letters */
  const char* const short_options = "p:h:s:t:?";

  /* valid long options */
  const struct option long_options[] = {
      { "port",		1, NULL, 'p' },
      { "hostname",	1, NULL, 'h' },
      { "sensorname",   1, NULL, 's' },      
      { "time intervall",0, NULL, 't' },
      { "help",		0, NULL, '?' },
      { NULL,		0, NULL,  0  }
  };

  *pHostname = NULL;
  *pPort = -1;
  *pSensorname = NULL;

  while( (next_option = getopt_long (argc, argv,
                        short_options, long_options, NULL)) != -1 )
  {

    switch( next_option )
    {
      // port
      case 'p':
        *pPort = atoi(optarg);
        break;
      // hostname
      case 'h':
        *pHostname = optarg;
        break;
      case 's':
        *pSensorname = optarg;
        break;
      case 't':
        *pIntervall = atoi(optarg);
        break;
      case '?':
        fprintf(stderr,"usage: %s -h <host> -p <port> -s <sensor name> [-t <sec>]\n", argv[0]);
        exit(0);
      default:
        fprintf(stderr, "Invalid option %c! \n", next_option);
        exit(0);
    }
  }
}

/* --------------------------------*/
void release_usb_device(int dummy)
{
      int ret;
      ret = usb_release_interface(devh, 0);
      usb_close(devh);
      exit(ret);
}

/* -----------------------------------------------------*/
struct usb_device* find_device(int vendor, int product)
{
      struct usb_bus *bus;

      for (bus = usb_get_busses(); bus; bus = bus->next)
      {
          struct usb_device *dev;
       
          for (dev = bus->devices; dev; dev = dev->next)
          {
              if (dev->descriptor.idVendor == vendor
                  && dev->descriptor.idProduct == product)
                return dev;
          }
      }
      return NULL;
}


struct sockaddr_in serverAddr;
struct hostent *server;
int sockfd, port;
char *hostname;
char *sensorname;

/* ------------------------------*/
void connect2server()
{
     // create a INET socket.
     if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
     {
         perror("socket failed");
         exit(0);
     }
     
     // initialize address
     bzero((char *) &serverAddr, sizeof(serverAddr));
     // protocol family is INET
     serverAddr.sin_family = AF_INET;
     // copy host address to structure
      bcopy((char *)server->h_addr,
        (char *)&serverAddr.sin_addr.s_addr,
        server->h_length);
      // port number must be in network byte order
      serverAddr.sin_port = htons(port);
      // now try to connect to server ...
      while (connect(sockfd,
         (struct sockaddr *) &serverAddr,
         sizeof(serverAddr)) < 0)
      {
         perror("Connecting failed. Retrying ...");
         sleep (2);
      }

     printf("Connected to server %s, port %d\n", hostname,port);
}

/* ------------------------------*/
int send2socket(char *sendstr)
{
    int n;

    signal(SIGPIPE, SIG_IGN); // ignore SIGPIPE in case socket is closed while writing to it

    if( (n = write(sockfd,sendstr,strlen(sendstr))) < 0 )
    {
        perror("write failed");
        return(0);
    }
    else
    {
        // reset buffer content
        bzero(sendstr,sizeof(sendstr));
        return(n);
    }
}

/* ------------------------------*/
int main(int argc, char *argv[])
{
    char buffer[100];
    int i, n;
    int ret, vendor, product, counter;
    struct usb_device *dev;
    char buf[1000];
    int intervall=60; // time between intermediate message to server (default:60 sec)

    vendor = 0x03eb;
    product = 0x2013;
    dev = NULL;

    if (argc < 3) 
    {
       fprintf(stderr,"usage1: %s -h <host> -p <port> -s <sensor name> [-t <secs>]\n", argv[0]);
       exit(0);
    }

    // retrieve given arguments ...
    getArguments( argc, argv, &port, &hostname, &sensorname, &intervall);

    // hostname and port must be set!
    // getArguments sets hostname to point to NULL
    // and port to -1. So we can detect, whether arguments
    // are given or not.

    if( hostname == NULL || port == -1 || sensorname == NULL)
    {
        fprintf(stderr,"usage2: %s -h <host> -p <port> -s <sensor name> [-t <secs>]\n", argv[0]);
        exit(1);
    }

    printf("Searching for USB device ...\n");
    usb_init();
    counter = 0;

    do
    {
       usb_set_debug(0);
       usb_find_busses();
       usb_find_devices();
       dev = find_device(vendor, product);
       sleep(10); 

       ++counter;

       if (counter == 10)
       {
            printf("Error: Device not found\n");
            exit(1);
       }

    }  while (dev == NULL);

    assert(dev);

    devh = usb_open(dev);
    assert(devh);

    signal(SIGTERM, release_usb_device);

    ret = usb_get_driver_np(devh, 0, buf, sizeof(buf));
    if (ret == 0)
    {
        ret = usb_detach_kernel_driver_np(devh, 0);
    }

    ret = usb_claim_interface(devh, 0);

    if (ret != 0) {
        printf("Error: claim failed with error: %d", ret);
        exit(1);
    }

    unsigned short iresult=0;
    unsigned short voc=0;

    ret = usb_interrupt_read(devh, 0x00000081, buf, 0x0000010, 1000);

    printf("USB device opened\n");

    if( (server = gethostbyname(hostname)) == NULL )
    {
       fprintf(stderr,"host %s not found!\n", hostname);
       exit;
    }

    connect2server();

// ######## MAIN LOOP ################

    while(0==0)
    {
         time_t t = time(NULL);
         struct tm tm = *localtime(&t);

         // USB COMMAND TO REQUEST DATA
         // @h*TR

         memcpy(buf, "\x40\x68\x2a\x54\x52\x0a\x40\x40\x40\x40\x40\x40\x40\x40\x40\x40", 0x0000010);
         ret = usb_interrupt_write(devh, 0x00000002, buf, 0x0000010, 1000);
         ret = usb_interrupt_read(devh, 0x00000081, buf, 0x0000010, 1000);

         if ( !((ret == 0) || (ret == 16)))
         {
             printf("ERROR: Invalid result code: %d\n", ret);
             continue;
         }
         if (ret == 0)
         {
             sleep(1);
             ret = usb_interrupt_read(devh, 0x00000081, buf, 0x0000010, 1000);
         }

         memcpy(&iresult,buf+2,2);
         voc = __le16_to_cpu(iresult);
         sleep(1);

         ret = usb_interrupt_read(devh, 0x00000081, buf, 0x0000010, 1000);

         printf("%s|%d\n",sensorname,voc);

         if ( voc >= 450 && voc <= 10000)
         {
            sprintf(buffer,"%s|%d",sensorname,voc);

            if(!send2socket(buffer))
            {
                 // something went wrong -> try to reconnect
                 close(sockfd);
                 printf("retrying to connect to host %s ...\n", hostname);
                 sleep(5);
                 connect2server();
            }
         }
         // Wait for next request for data
         sleep(intervall);
     }
     close(sockfd);
}

